/*-
 * Copyright (c) 2001 Allan Saddi <asaddi@philosophysw.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: ruleset.c,v 1.11 2001/08/16 20:20:43 asaddi Exp $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <ctype.h>
#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "yafic.h"

#ifndef lint
static const char rcsid[] =
	"$Id: ruleset.c,v 1.11 2001/08/16 20:20:43 asaddi Exp $";
#endif /* !lint */

#define LINE_BUFFER_SIZE 4096

static const char flagChars[] = "pinugsamch";
static const struct {
  char c;
  rflag_t flags;
} flagTemplate[] = {
  { 'R', RFLAG_DEFAULT },
  { 'L', RFLAG_DEFAULT & ~(RFLAG_MTIME | RFLAG_CTIME | RFLAG_HASH) },
  { 'N', RFLAG_DEFAULT | RFLAG_SIZE | RFLAG_ATIME },
  { 'E', 0 }
};
#define TEMPLATE_COUNT (sizeof (flagTemplate) / sizeof (flagTemplate[0]))

static struct RuleEntry **ruleSet;

void
ApplyRuleSet (void (*func) (struct RuleEntry *re))
{
  int i;
  struct RuleEntry *re;

  for (i = RULESET_TABLE_SIZE - 1; i >= 0; i--) {
    re = ruleSet[i];
    while (re) {
      func (re);
      re = re->next;
    }
  }
}

static void
destroyRuleEntry (struct RuleEntry *re)
{
  free (re->path);
  free (re);
}

struct RuleEntry *
FindRuleEntry (const char *path)
{
  struct RuleEntry *re;
  uint32_t hash;

  hash = myhash (path) % RULESET_TABLE_SIZE;
  re = ruleSet[hash];
  while (re) {
    if (!strcmp (re->path, path))
      return re;
    re = re->next;
  }

  return NULL;
}

struct RuleEntry *
FindClosestRuleEntry (const char *path)
{
  char *newpath;
  struct RuleEntry *re;
  char *tmp;

  newpath = mystrdup (path);
  /* There better be a root entry or else we're going to loop forever! */
  for (;;) {
    if ((re = FindRuleEntry (newpath))) {
      free (newpath);
      return re;
    }
    tmp = mydirname (newpath);
    free (newpath);
    newpath = tmp;
  }
  /* NOTREACHED */
}

struct RuleEntry *
addRuleEntry (const char *path, rflag_t entryFlags, rflag_t descFlags)
{
  struct RuleEntry *re;
  uint32_t hash;

  re = mymalloc (sizeof (*re));
  re->path = mystrdup (path);
  re->entryFlags = entryFlags;
  re->descFlags = descFlags;

  hash = myhash (re->path) % RULESET_TABLE_SIZE;
  re->next = ruleSet[hash];
  ruleSet[hash] = re;

  return re;
}

void
InitRuleSet (void)
{
  /* Initialize hash table. */
  ruleSet = mymalloc (sizeof (*ruleSet) * RULESET_TABLE_SIZE);
  memset (ruleSet, 0, sizeof (*ruleSet) * RULESET_TABLE_SIZE);

  /* Populate with root entry. */
  addRuleEntry ("/", RFLAG_DEFAULT, RFLAG_DEFAULT);
}

void
CleanRuleSet (void)
{
  int i;
  struct RuleEntry *re, *nextre;

  if (!ruleSet)
    return;

  for (i = RULESET_TABLE_SIZE - 1; i >= 0; i--) {
    re = ruleSet[i];
    while (re) {
      nextre = re->next;
      destroyRuleEntry (re);
      re = nextre;
    }
  }

  free (ruleSet);
}

static int
parseFlags (const char *flagsStr, rflag_t *flags)
{
  int i;
  int addMode = 1;

  if (flagsStr[0] == '+' || flagsStr[0] == '-') {
    *flags = RFLAG_DEFAULT;
  }
  else {
    *flags = 0;

    for (i = 0; i < TEMPLATE_COUNT; i++) {
      if (flagTemplate[i].c == *flagsStr) {
	*flags = flagTemplate[i].flags;
	flagsStr++;
	addMode = -1;
	break;
      }
    }
  }

  while (*flagsStr) {
    if (*flagsStr == '+') {
      addMode = 1;
      flagsStr++;
      continue;
    }
    else if (*flagsStr == '-') {
      addMode = 0;
      flagsStr++;
      continue;
    }

    for (i = 0; i < RFLAG_MAX; i++)
      if (*flagsStr == flagChars[i])
	break;

    if (i < RFLAG_MAX) {
      if (addMode == -1)
	return 0;
      else if (addMode)
	*flags |= 1 << i;
      else
	*flags &= ~(1 << i);
    }
    else
      return 0;

    flagsStr++;
  }

  return 1;
}

static int
checkPath (const char *path)
{
  /* Make sure there are no empty path components. */
  if (strstr (path, "//"))
    return 0;

  /* Make sure it's an absolute path. */
  if (path[0] != '/')
    return 0;

  /* Looks ok. */
  return 1;
}

static void
resolveRuleEntry (struct RuleEntry *re)
{
  struct RuleEntry *parentre;
  char *parent, *tmp;

  if (re->descFlags & RFLAG_UPDATE) {
    parent = mydirname (re->path);

    for (;;) {
      if ((parentre = FindRuleEntry (parent)) &&
	  !(parentre->descFlags & RFLAG_IGNORE)) {
	re->descFlags = parentre->descFlags;
	break;
      }
      if (!strcmp (parent, "/")) {
	/* End of the line, use default. */
	re->descFlags = RFLAG_DEFAULT;
	break;
      }
      /* Try next ancestor. */
      tmp = mydirname (parent);
      free (parent);
      parent = tmp;
    }

    free (parent);
  }
}

int
ParseRuleSet (const char *conf)
{
  FILE *f;
  int success = 1;
  char *line, *comment, buf[LINE_BUFFER_SIZE];
  int lineNo;
  int i, len;
  char *entry, *flagsStr;
  rflag_t flags;
  struct RuleEntry *re, *nextre, *lastre;

  if ((f = fopen (conf, "r"))) {
    lineNo = 0;
    while (fgets (buf, sizeof (buf), f)) {
      line = buf;
      lineNo++;

      /* Strip leading whitespace. */
      while (*line && isspace ((int) *line))
	line++;

      /* Strip comments. */
      if ((comment = strchr (line, '#')))
	*comment = '\0';

      /* Skip empty lines. */
      if (!(*line))
	continue;

      /* Strip trailing whitespace. */
      i = strlen (line) - 1;
      while (i >= 0 && isspace ((int) line[i]))
	line[i--] = '\0';

      /* Grab entry and flags. */
      entry = strtok (line, " \t");
      flagsStr = strtok (NULL, " \t");
      if (strtok (NULL, " \t"))
	fprintf (stderr, "%s: %s:%d: ignoring junk after flags\n",
		 prog, conf, lineNo);

      /* Wipe out trailing slash... except if it's just '/'. */
      i = strlen (entry) - 1;
      if (i > 1 && entry[i] == '/')
	entry[i] = '\0';

      /* Make sure entry is correctly formed. */
      if (entry[0] == '!' || entry[0] == '=' || entry[0] == '$') {
	if (!checkPath (&entry[1])) {
	  fprintf (stderr, "%s: %s:%d: bad path\n",
		   prog, conf, lineNo);
	  success = 0;
	  continue;
	}
      }
      else if (!checkPath (entry)) {
	fprintf (stderr, "%s: %s:%d: bad path\n",
		 prog, conf, lineNo);
	success = 0;
	continue;
      }

      if (!flagsStr)
	flags = RFLAG_DEFAULT;
      else if (!parseFlags (flagsStr, &flags)) {
	fprintf (stderr, "%s: %s:%d: bad flags\n",
		 prog, conf, lineNo);
	success = 0;
	continue;
      }

      /* Parse the entry. */
      if (*entry == '!') {
	/* Ignore entry. */
	entry++;
	/* Wipe out existing entries. */
	len = strlen (entry);
	for (i = 0; i < RULESET_TABLE_SIZE; i++) {
	  lastre = NULL;
	  re = ruleSet[i];
	  while (re) {
	    nextre = re->next;
	    if (!strncmp (entry, re->path, len)) {
	      if (!lastre)
		ruleSet[i] = nextre;
	      else
		lastre->next = nextre;
	      destroyRuleEntry (re);
	    }
	    else
	      lastre = re;
	    re = nextre;
	  }
	}
	addRuleEntry (entry, RFLAG_IGNORE, RFLAG_IGNORE);
      }
      else if (*entry == '=') {
	/* Directory only, no recursion. */
	entry++;
	if ((re = FindRuleEntry (entry))) {
	  re->entryFlags = flags;
	  re->descFlags = RFLAG_IGNORE;
	}
	else
	  addRuleEntry (entry, flags, RFLAG_IGNORE);
      }
      else if (*entry == '$') {
	/* Directory only, recurse. */
	entry++;
	if ((re = FindRuleEntry (entry)))
	  re->entryFlags = flags;
	else
	  addRuleEntry (entry, flags, RFLAG_UPDATE);
      }
      else {
	/* Normal entry. */
	if ((re = FindRuleEntry (entry))) {
	  re->entryFlags = flags;
	  re->descFlags = flags;
	}
	else
	  addRuleEntry (entry, flags, flags);
      }
    }
    if (ferror (f))
      yaficError (conf);
  }
  else
    yaficError (conf);

  if (success)
    ApplyRuleSet (resolveRuleEntry);

  return success;
}

static char *
dumpFlags (rflag_t flags)
{
  char *p;
  int i;
  rflag_t test;
  static char buf[16];

  if (flags & RFLAG_IGNORE) {
    buf[0] = '-';
    buf[1] = '\0';
    return buf;
  }

  p = buf;
  for (i = 0, test = 1; i < RFLAG_MAX; i++, test <<= 1)
    if (flags & test)
      *(p++) = flagChars[i];
  *(p++) = '\0';

  return buf;
}

void
DumpRuleEntry (struct RuleEntry *re)
{
  printf ("  %s (%s, ", re->path, dumpFlags (re->entryFlags));
  printf ("%s)\n", dumpFlags (re->descFlags));
}
